///////////////////////////////////////////////////////////////////////////////
//
// Filename: FTDiagramFrame.java
//
// Author:   Robert Thornburrow
// Project:  FamilyTree
// Created:  20th May 2001
//
// Purpose:  Window for a Family Tree Diagram.
//
// (c) Copyright R.H.Thornburrow, 2001
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
//
///////////////////////////////////////////////////////////////////////////////

package com.tsac.familytree.diagram;

import com.tsac.familytree.*;
import com.tsac.familytree.gui.PageSetupDialog;

import java.awt.BorderLayout;
import java.awt.GridBagLayout;
import java.awt.Dimension;
import java.awt.Point;
import java.awt.Insets;
import java.awt.PrintJob;
import java.awt.Graphics;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.Cursor;

import javax.swing.JScrollPane;
import javax.swing.ScrollPaneConstants;
import javax.swing.JMenuBar;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import java.awt.*;
import java.util.ResourceBundle;

public class FTDiagramFrame
    extends javax.swing.JFrame
{
	static ResourceBundle res1 = ResourceBundle.getBundle("FamilyTree");
	
    public FTDiagramFrame( Person root )
    {    
		//{{INIT_CONTROLS
		setJMenuBar(menubar);
		getContentPane().setLayout(new BorderLayout(0,0));
		getContentPane().setBackground(java.awt.Color.white);
		setSize(800,600);
		setVisible(false);
		diagscroll.setAutoscrolls(true);
		getContentPane().add(BorderLayout.CENTER,diagscroll);
		diagscroll.getViewport().setBackground(java.awt.Color.white);
		diagscroll.setBounds(0,0,800,600);
		diagscroll.getViewport().add(diagram);
		diagram.setBackground(java.awt.Color.white);
		diagram.setBounds(0,0,797,597);
		//$$ menubar.move(0,601);
		menubar.add(fileMenu);
		fileMenu.add(filePageSetupBtn);
		fileMenu.add(filePrintBtn);
		fileMenu.add(fileCloseBtn);
		//}}

		fileMenu.setText(res1.getString("File"));
		filePageSetupBtn.setText(res1.getString("Page_Setup_"));
		filePrintBtn.setText(res1.getString("Print_"));
		fileCloseBtn.setText(res1.getString("Close"));

		filePageSetupBtn.addActionListener( new ActionListener() {
		    public void actionPerformed( final ActionEvent e ) {
		        actionPageSetup();
		    }
		});

		filePrintBtn.addActionListener( new ActionListener() {
		    public void actionPerformed( final ActionEvent e ) {
		        actionPrint();
		    }
		});

		fileCloseBtn.addActionListener( new ActionListener() {
		    public void actionPerformed( final ActionEvent e ) {
		        actionClose();
		    }
		});

		diagram.setRoot( root );
		this.setTitle( root.getSurname() + res1.getString("Family_Tree") );
	}
	//{{DECLARE_CONTROLS
    javax.swing.JScrollPane diagscroll = new javax.swing.JScrollPane();
	com.tsac.familytree.diagram.FTDiagram diagram = new com.tsac.familytree.diagram.FTDiagram();
	javax.swing.JMenuBar menubar = new javax.swing.JMenuBar();
	javax.swing.JMenu fileMenu = new javax.swing.JMenu();
	javax.swing.JMenuItem filePageSetupBtn = new javax.swing.JMenuItem();
	javax.swing.JMenuItem filePrintBtn = new javax.swing.JMenuItem();
	javax.swing.JMenuItem fileCloseBtn = new javax.swing.JMenuItem();
	//}}
	//{{DECLARE_MENUS
	//}}


	private void actionPageSetup()
	{
	    new PageSetupDialog( this, true ).setVisible( true );
	}


	private void actionPrint()
	{
        String title = diagram.getSurname() + res1.getString("Family_Tree");
        PrintJob job = Toolkit.getDefaultToolkit().getPrintJob( this, title, null );

        // Check if cancel was pressed
        if( job == null ) return;

        Dimension psz = job.getPageDimension();
        int pageWidth  = psz.width - (PageSetupDialog.leftMargin + PageSetupDialog.rightMargin);
        int pageHeight = psz.height - (PageSetupDialog.topMargin + PageSetupDialog.bottomMargin);

        if( PageSetupDialog.multipage )
        {
            int width = PageSetupDialog.multipageWidth;
            int height = PageSetupDialog.multipageHeight;
            Dimension size = new Dimension( pageWidth * width, pageHeight * height );
	        setCursor( new Cursor( Cursor.WAIT_CURSOR ) );

            for( int x = 0; x < width; x++ )
            {
                for( int y = 0; y < height; y++ )
                {
                    Graphics jg = job.getGraphics();

                    try
                    {
                        Graphics pg = jg.create( PageSetupDialog.leftMargin, PageSetupDialog.topMargin, pageWidth, pageHeight );
                        Point origin = new Point( -(pageWidth * x), -(pageHeight * y) );

                        diagram.print( pg, origin, size );
                        pg.dispose();
                    }
                    catch( Exception e )
                    {
                        e.printStackTrace();
                        System.out.println( e );
                        break;
                    }
                    finally
                    {
                        jg.dispose();
                    }
                } // y loop
            } // x loop

            job.end();
            job = null;
	        setCursor( Cursor.getDefaultCursor() );
	    }
	    else
	    {
	        setCursor( new Cursor( Cursor.WAIT_CURSOR ) );
            Graphics jg = job.getGraphics();

            try
            {
                Graphics pg = jg.create( PageSetupDialog.leftMargin, PageSetupDialog.topMargin, pageWidth, pageHeight );
                Dimension size = new Dimension( pageWidth, pageHeight );

                diagram.print( pg, new Point(0,0), size );
                pg.dispose();
            }
            catch( Exception e )
            {
                e.printStackTrace();
                System.out.println( e );
            }
            finally
            {
                jg.dispose();
                job.end();
                job = null;
    	        setCursor( Cursor.getDefaultCursor() );
            }
	    }
	}


	private void actionClose()
	{
	    this.setVisible( false );
	}
}