///////////////////////////////////////////////////////////////////////////////
//
// Filename: MainFrame.java
//
// Author:   Robert Thornburrow
// Project:  FamilyTree
// Created:  20th May 2001
//
// Purpose:  The Main Window.
//
// (c) Copyright R.H.Thornburrow, 2001
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
//
///////////////////////////////////////////////////////////////////////////////
//
// Contributors: Janne Ilonen <janne.ilonen@virpi.net> Added *.properties files
// for multi language supports. Used JBuilder5 Wizards - Resource Strings...
// Date 2002-09-04
///////////////////////////////////////////////////////////////////////////////

package com.tsac.familytree.gui;

import com.tsac.familytree.*;
import com.tsac.familytree.util.*;
import com.tsac.familytree.csv.CsvLoader;
import com.tsac.familytree.gedcom.GedcomLoader;
import com.tsac.familytree.ftml.FtmlLoader;
import com.tsac.familytree.diagram.FTDiagramFrame;
import com.tsac.familytree.ancestor.AncestorFrame;
import com.tsac.familytree.timeline.TLDiagramFrame;

import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.BorderLayout;
import java.awt.Insets;
import java.awt.Dimension;
import java.awt.Color;
import java.awt.Cursor;
import java.awt.Graphics;
import java.awt.PrintJob;
import java.awt.Toolkit;
import java.awt.event.*;

import javax.swing.*;
import javax.swing.tree.*;
import javax.swing.event.*;

import java.util.Vector;
import java.util.Enumeration;
import java.io.StringWriter;
import java.io.PrintWriter;
import java.net.URL;
import java.util.ResourceBundle;


public class MainFrame
    extends javax.swing.JFrame
    implements Runnable
{
    static ResourceBundle res = ResourceBundle.getBundle("FamilyTree");
    public static String version = "1.1.2";

    private FamilyTreeModel _model;
    private FamilyTreeTableModel _tableModel = new FamilyTreeTableModel();
    private Person _root;
    private Person[] _rootsConventional;
    private Person[] _rootsInverted;
    private Person[] _people;

    private String _filename = "treedata.xml";

    private FamilyTreeNode _lastSelectedNode = null;
    private int _showFlags = SHOW_ALL;
    private int _orientation = FamilyTreeModel.INVERTED;
    private int _selectedRoot = 0;
    private int _selectedInvRoot = 0;

    private Vector _history = new Vector();

    public static final int SHOW_PARENTS     = 0x01;
    public static final int SHOW_SPOUSES     = 0x02;
    public static final int SHOW_CHILDREN    = 0x04;
    public static final int SHOW_SIBLINGS    = 0x08;
    public static final int SHOW_AUNTSUNCLES = 0x10;
    public static final int SHOW_COUSINS     = 0x20;
    public static final int SHOW_ALL         = 0xff;

	public MainFrame()
	{
	    tree.setModel( new FamilyTreeModel() );

		// This code is automatically generated by Visual Cafe when you add
		// components to the visual environment. It instantiates and initializes
		// the components. To modify the code, only use code syntax that matches
		// what Visual Cafe can generate, or Visual Cafe may be unable to back
		// parse your Java file into its visual environment.
		//{{##INIT_CONTROLS
		setJMenuBar(menubar);
		getContentPane().setLayout(new GridBagLayout());
		setSize(800,600);
		setVisible(false);
		getContentPane().add(horizsplit, new FamilyTreeGridBagConstraints(0,0,1,1,1.0,1.0,GridBagConstraints.CENTER,GridBagConstraints.BOTH,625,431));
		horizsplit.setBounds(0,0,625,414);
		treescrollpane.setAutoscrolls(true);
		treescrollpane.setOpaque(true);
		horizsplit.add(treescrollpane,JSplitPane.LEFT);
		treescrollpane.setBounds(1,1,250,431);
		tree.setShowsRootHandles(true);
		tree.setRowHeight(16);
		tree.setAutoscrolls(true);
		treescrollpane.getViewport().add(tree);
		tree.setBounds(0,0,247,428);
		vertsplit.setOrientation(javax.swing.JSplitPane.VERTICAL_SPLIT);
		horizsplit.add(vertsplit,JSplitPane.RIGHT);
		vertsplit.setBounds(0,0,375,414);
		tablescrollpane.setOpaque(true);
		vertsplit.add(tablescrollpane,JSplitPane.TOP);
		tablescrollpane.setBounds(0,0,375,175);
		table.setShowHorizontalLines(false);
		tablescrollpane.getViewport().add(table);
		table.setBounds(0,0,375,175);
		docscrollpane.setHorizontalScrollBarPolicy(javax.swing.ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);
		docscrollpane.setOpaque(true);
		vertsplit.add(docscrollpane,JSplitPane.BOTTOM);
		docscrollpane.setBounds(0,0,20,40);
		doc.setEditable(false);
		docscrollpane.getViewport().add(doc);
		doc.setBounds(0,0,448,37);
		statusTxt.setEditable(false);
		getContentPane().add(statusTxt, new FamilyTreeGridBagConstraints(0,1,1,1,1.0,0.0,GridBagConstraints.CENTER,GridBagConstraints.HORIZONTAL,625,2));
		statusTxt.setBounds(0,414,625,19);
		//$$ menubar.move(48,434);
		fileMenu.setText(res.getString("File"));
		fileMenu.setActionCommand(res.getString("JMenu1"));
		menubar.add(fileMenu);
		fileOpenBtn.setText(res.getString("Open_"));
		fileMenu.add(fileOpenBtn);
		filePageSetupBtn.setText(res.getString("Page_Setup_"));
		fileMenu.add(filePageSetupBtn);
		fileExitBtn.setText(res.getString("Exit"));
		fileExitBtn.setActionCommand(res.getString("JMenuItem1"));
		fileMenu.add(fileExitBtn);
		viewMenu.setText(res.getString("View"));
		viewMenu.setActionCommand(res.getString("View"));
		menubar.add(viewMenu);
		showMenu.setText(res.getString("Show"));
		showMenu.setActionCommand(res.getString("Show"));
		viewMenu.add(showMenu);
		showParentsChk.setText(res.getString("Parents"));
		showParentsChk.setActionCommand(res.getString("Parents"));
		showMenu.add(showParentsChk);
		showChildrenChk.setText(res.getString("Children"));
		showChildrenChk.setActionCommand(res.getString("Children"));
		showMenu.add(showChildrenChk);
		showSiblingsChk.setText(res.getString("Siblings"));
		showSiblingsChk.setActionCommand(res.getString("Siblings"));
		showMenu.add(showSiblingsChk);
		showAuntsUnclesChk.setText(res.getString("Aunts_Uncles"));
		showAuntsUnclesChk.setActionCommand(res.getString("Aunts_Uncles"));
		showMenu.add(showAuntsUnclesChk);
		showSpousesChk.setText(res.getString("Spouses"));
		showMenu.add(showSpousesChk);
		showCousinsChk.setText(res.getString("Cousins"));
		showCousinsChk.setActionCommand(res.getString("Cousins"));
		showMenu.add(showCousinsChk);
		orientMenu.setText(res.getString("Orientation"));
		orientMenu.setActionCommand(res.getString("Orientation"));
		viewMenu.add(orientMenu);
		orientNormalRb.setText(res.getString("Normal"));
		orientNormalRb.setActionCommand(res.getString("Normal"));
		orientMenu.add(orientNormalRb);
		orientInvertedRb.setText(res.getString("Inverted"));
		orientInvertedRb.setActionCommand(res.getString("Inverted"));
		orientMenu.add(orientInvertedRb);
		showSpousesChk.setActionCommand(res.getString("Spouses"));
		viewSetRootMenu.setText(res.getString("Set_Root"));
		viewMenu.add(viewSetRootMenu);
		toolsMenu.setText(res.getString("Tools"));
		toolsMenu.setActionCommand(res.getString("Tools"));
		menubar.add(toolsMenu);
		toolsDrawTreeBtn.setText(res.getString("Draw_Tree"));
		toolsDrawTreeBtn.setActionCommand(res.getString("Draw_Tree"));
		toolsMenu.add(toolsDrawTreeBtn);
		toolsDrawAncestorsBtn.setText(res.getString("Draw_Ancestors"));
		toolsDrawAncestorsBtn.setActionCommand(res.getString("Draw_Ancestors"));
		toolsMenu.add(toolsDrawAncestorsBtn);
		toolsDrawTimelineBtn.setText(res.getString("Draw_Timeline"));
		toolsDrawTimelineBtn.setActionCommand(res.getString("Draw_Timeline"));
		toolsMenu.add(toolsDrawTimelineBtn);
		helpMenu.setText(res.getString("Help"));
		helpMenu.setActionCommand(res.getString("Help"));
		menubar.add(helpMenu);
		helpAboutBtn.setText(res.getString("About_"));
		helpAboutBtn.setActionCommand(res.getString("About_"));
		helpMenu.add(helpAboutBtn);
		//$$ JMenuBar1.move(48,434);
		//##}}

		//{{INIT_MENUS
		//}}

		//{{REGISTER_LISTENERS
		SymAction lSymAction = new SymAction();
		fileExitBtn.addActionListener(lSymAction);
		//}}

		// Build doc popup menu
        _docPopupMenu.add( _docPopupBackBtn );
        _docPopupMenu.add( _docPopupPrintBtn );

		this.setTitle( "Taurus FamilyTree" );
		table.setModel(_tableModel);

		statusTxt.setBackground( Color.lightGray );
		tablescrollpane.setBackground( Color.white );
		tree.setCellRenderer( new FamilyTreeNodeRenderer() );

		// Add a listener to the tree to allows list and doc updates
		tree.addTreeSelectionListener( new TreeSelectionListener() {
		    public void valueChanged( TreeSelectionEvent e ) {
		        Object sel = e.getPath().getLastPathComponent();
		        if( sel != null ) actionTreeNodeSelected( (FamilyTreeNode) sel );
		    }
		});

		// add a listener to give tree a popup menu
		tree.addMouseListener( new MouseAdapter() {
		    public void mouseClicked( MouseEvent e ) {
		        if( (e.getModifiers() & InputEvent.BUTTON3_MASK) != 0 )
		            actionTreePopup( e.getX(), e.getY() );
		    }
		} );

		// Add a listener to the list to allow doc updates
		table.getSelectionModel().addListSelectionListener( new ListSelectionListener() {
		    public void valueChanged( ListSelectionEvent e ) {
		        if( !e.getValueIsAdjusting() )
		            actionListRowSelected( table.getSelectedRow() );
		    }
		});

		// Add a listener to make our hyperlinks do something
		doc.addHyperlinkListener( new HyperlinkListener() {
		    public void hyperlinkUpdate( HyperlinkEvent e ) {
		        if( e.getEventType() == HyperlinkEvent.EventType.ACTIVATED )
		            actionHyperlinkActivated( e.getURL() );
		        else if( e.getEventType() == HyperlinkEvent.EventType.ENTERED )
		            statusTxt.setText( e.getURL().toString() );
		        else if( e.getEventType() == HyperlinkEvent.EventType.EXITED )
		            statusTxt.setText( "" );
		    }
		} );

		// add a listener to give doc a popup menu
		doc.addMouseListener( new MouseAdapter() {
		    public void mouseClicked( MouseEvent e ) {
		        if( (e.getModifiers() & InputEvent.BUTTON3_MASK) != 0 )
		            actionDocumentPopup( e.getX(), e.getY() );
		    }
		} );

		// Add window close listener
		this.addWindowListener( new WindowAdapter() {
		    public void windowClosing( WindowEvent e ) {
		        System.exit(0);
		    }
		} );

		// Setup the scrollbars for the tree
		treescrollpane.setMinimumSize( tree.getPreferredScrollableViewportSize() );
		treescrollpane.setMaximumSize( new Dimension( 65536, 65536 ) );

		// Move the splitter bars
		horizsplit.setDividerLocation( 0.33 );
		vertsplit.setDividerLocation( 0.4 );

		// Menu handling code
		fileOpenBtn.addActionListener( new ActionListener() {
		    public void actionPerformed( final ActionEvent event ) {
		        actionOpen();
		    }
		});

		filePageSetupBtn.addActionListener( new ActionListener() {
		    public void actionPerformed( final ActionEvent event ) {
		        actionPageSetup();
		    }
		});

        ActionListener showAction = new ActionListener() {
		    public void actionPerformed( final ActionEvent event ) {
		        actionShowChanged();
		    }
	    };
        showParentsChk.addActionListener( showAction );
        showSpousesChk.addActionListener( showAction );
        showChildrenChk.addActionListener( showAction );
        showSiblingsChk.addActionListener( showAction );
        showAuntsUnclesChk.addActionListener( showAction );
        showCousinsChk.addActionListener( showAction );

        ActionListener orientAction = new ActionListener() {
		    public void actionPerformed( final ActionEvent event ) {
		        actionOrientChanged();
		    }
	    };
        orientNormalRb.addActionListener( orientAction );
        orientInvertedRb.addActionListener( orientAction );

        toolsDrawTreeBtn.addActionListener( new ActionListener() {
		    public void actionPerformed( final ActionEvent event ) {
		        actionDrawTree();
		    }
        } );

        toolsDrawAncestorsBtn.addActionListener( new ActionListener() {
		    public void actionPerformed( final ActionEvent event ) {
		        actionDrawAncestors();
		    }
        } );

        toolsDrawTimelineBtn.addActionListener( new ActionListener() {
		    public void actionPerformed( final ActionEvent event ) {
		        actionDrawTimeline();
		    }
        } );

        helpAboutBtn.addActionListener( new ActionListener() {
		    public void actionPerformed( final ActionEvent event ) {
		        new AboutDialog( MainFrame.this, true ).show();
		    }
        } );

        ButtonGroup orientGrp = new ButtonGroup();
        orientGrp.add( orientNormalRb );
        orientGrp.add( orientInvertedRb );

        buildSetRootMenu( _rootsInverted );

        updateScreenState();

        // Handle history back menu button
        _docPopupBackBtn.addActionListener( new ActionListener() {
            public void actionPerformed( final ActionEvent event ) {
                actionHistoryBack();
            }
        } );

        // Handle document print menu button
        _docPopupPrintBtn.addActionListener( new ActionListener() {
            public void actionPerformed( final ActionEvent event ) {
                actionDocumentPrint();
            }
        } );
    }


	public void run()
	{
	    String extn = "";
	    FamilyTreeLoader loader;

	    if( _filename.indexOf( '.' ) != -1 )
	        extn = _filename.substring( _filename.indexOf( '.' ) );

	    if( extn.equals( ".csv" ) )
	        loader = new CsvLoader();
	    else if( extn.equals( ".ged" ) )
	        loader = new GedcomLoader();
	    else
	        loader = new FtmlLoader();

        _root = null;
        _rootsConventional = null;
        _rootsInverted = null;
        _people = null;
        Person.clearKeys();
        _lastSelectedNode = null;
	    tree.setModel( new FamilyTreeModel() );

	    setBusy( res.getString("Loading") + " " + _filename + "'..." );

	    long start = System.currentTimeMillis();
	    loader.load( _filename );
	    long end = System.currentTimeMillis();

	    System.out.println( res.getString("Loaded") + " " + _filename + " " + res.getString("in") + " " + (end - start) + " ms" );
        setBusy( null );

	    populateModel( loader );
		tree.setModel( _model );

        buildSetRootMenu( (_orientation == FamilyTreeModel.CONVENTIONAL) ? _rootsConventional : _rootsInverted );
        updateScreenState();

        actionTreeNodeSelected( (FamilyTreeNode) tree.getModel().getRoot() );
	}


	public void setFilename( String filename )  { _filename = filename; }


    public void populateModel( FamilyTreeLoader loader  )
    {
        Marriage[] marriages = loader.getMarriages();
        _people = loader.getPeople();
	    _rootsInverted = loader.getInvertedRoots();
	    _rootsConventional = loader.getConventionalRoots();

		// Calculate children and siblings
		for( int n = 0; n < _people.length; n++ )
		{
		    // This node is a sibling of those with the same father and/or mother
		    for( int m = 0; m < _people.length; m++ )
		    {
		        if( m != n )
		        {
		            if( (_people[n].father != null) && (_people[m].father == _people[n].father ) )
		                _people[n].addSibling( _people[m] );
		            else if( (_people[n].mother != null) && (_people[m].mother == _people[n].mother ) )
		                _people[n].addSibling( _people[m] );
		        }
		    }
		}

		// Add spouses from marriage table
		for( int n = 0; n < marriages.length; n++ )
		{
		    marriages[n].wife.addSpouse( marriages[n].husband );
		    marriages[n].husband.addSpouse( marriages[n].wife );
		}

		// Calculate inverted roots if none given
		if( (_rootsInverted == null) || (_rootsInverted.length == 0) )
		{
            Vector rootsInv = new Vector();

            for( int n = 0; n < _people.length; n++ )
            {
                if( (_people[n].children == null) || (_people[n].children.size() == 0) )
                    rootsInv.addElement( _people[n] );
            }

            _rootsInverted = new Person[rootsInv.size()];
            rootsInv.copyInto( _rootsInverted );
		}

		// Calculate normal roots if none given
		if( (_rootsConventional == null) || (_rootsConventional.length == 0) )
		{
            Vector rootsConv = new Vector();

            for( int n = 0; n < _people.length; n++ )
            {
                if( (_people[n].father == null) && (_people[n].mother == null ) )
                    rootsConv.addElement( _people[n] );
            }

            _rootsConventional = new Person[rootsConv.size()];
            rootsConv.copyInto( _rootsConventional );
		}

		// set selected roots
	    _selectedRoot = 0;
	    _selectedInvRoot = 0;

	    _root = (_orientation == FamilyTreeModel.CONVENTIONAL)
	            ? _rootsConventional[_selectedRoot]
	            : _rootsInverted[_selectedInvRoot];

        // create a new model
		_model = new FamilyTreeModel( _people, _root, _orientation );
	}

    private void actionTreeNodeSelected( FamilyTreeNode node )
    {
        _lastSelectedNode = node;

        setBusy( res.getString("Calculating_") );

        try
        {
            _tableModel.removeAllRows();

            _tableModel.addRow( new FamilyTreeTableRowModel( node.person, " ", " ", " " ) );

            // father
            if( isShowing(SHOW_PARENTS) && (node.person.father != null) )
                _tableModel.addRow( new FamilyTreeTableRowModel( node.person.father, res.getString("father"), res.getString("ERROR_"), res.getString("father") ) );

            // mother
            if( isShowing(SHOW_PARENTS) && (node.person.mother != null) )
                _tableModel.addRow( new FamilyTreeTableRowModel( node.person.mother, res.getString("ERROR_"), res.getString("mother"), res.getString("mother") ) );

            // husbands or wives
            if( isShowing(SHOW_SPOUSES) && (node.person.spouses != null) )
            {
                for( Enumeration e = node.person.spouses.elements(); e.hasMoreElements(); )
                    _tableModel.addRow( new FamilyTreeTableRowModel( (Person) e.nextElement(), res.getString("husband"), res.getString("wife"), res.getString("spouse") ) );
            }

            // sons and daughters
            if( isShowing(SHOW_CHILDREN) && (node.person.children != null) )
            {
                for( Enumeration e = node.person.children.elements(); e.hasMoreElements(); )
                    _tableModel.addRow( new FamilyTreeTableRowModel( (Person) e.nextElement(), res.getString("son"), res.getString("daughter"), res.getString("child") ) );
            }

            // brothers and sisters
            if( isShowing(SHOW_SIBLINGS) && (node.person.siblings != null) )
            {
                for( Enumeration e = node.person.siblings.elements(); e.hasMoreElements(); )
                    _tableModel.addRow( new FamilyTreeTableRowModel( (Person) e.nextElement(), res.getString("brother"), res.getString("sister"), res.getString("sibling") ) );
            }

            // aunts and uncles on fathers-side
            if( isShowing(SHOW_AUNTSUNCLES) && (node.person.father != null) && (node.person.father.siblings != null) )
            {
                for( Enumeration e = node.person.father.siblings.elements(); e.hasMoreElements(); )
                    _tableModel.addRow( new FamilyTreeTableRowModel( (Person) e.nextElement(), res.getString("uncle_father_"), res.getString("aunt_father_"), res.getString("uncle_aunt_father_") ) );
            }

            // aunts and uncles on mothers-side
            if( isShowing(SHOW_AUNTSUNCLES) && (node.person.mother != null) && (node.person.mother.siblings != null) )
            {
                for( Enumeration e = node.person.mother.siblings.elements(); e.hasMoreElements(); )
                    _tableModel.addRow( new FamilyTreeTableRowModel( (Person) e.nextElement(), res.getString("uncle_mother_"), res.getString("aunt_mother_"), res.getString("uncle_aunt_mother_") ) );
            }

            // cousins on fathers-side
            if( isShowing(SHOW_COUSINS) && (node.person.father != null) && (node.person.father.siblings != null) ) {
                for( Enumeration e1 = node.person.father.siblings.elements(); e1.hasMoreElements(); ) {
                    Person auntOrUncle = (Person) e1.nextElement();
                    if( auntOrUncle.children != null ) {
                        for( Enumeration e2 = auntOrUncle.children.elements(); e2.hasMoreElements(); )
                            _tableModel.addRow( new FamilyTreeTableRowModel( (Person) e2.nextElement(), res.getString("cousin_father_"), res.getString("cousin_father_"), res.getString("cousin_father_") ) );
                    }
                }
            }

            // cousins on mothers-side
            if( isShowing(SHOW_COUSINS) && (node.person.mother != null) && (node.person.mother.siblings != null) ) {
                for( Enumeration e1 = node.person.mother.siblings.elements(); e1.hasMoreElements(); ) {
                    Person auntOrUncle = (Person) e1.nextElement();
                    if( auntOrUncle.children != null ) {
                        for( Enumeration e2 = auntOrUncle.children.elements(); e2.hasMoreElements(); )
                            _tableModel.addRow( new FamilyTreeTableRowModel( (Person) e2.nextElement(), res.getString("cousin_mother_"), res.getString("cousin_mother_"), res.getString("cousin_mother_") ) );
                    }
                }
            }
        }
        catch( Exception e )
        {
            reportException( e );
        }

        try
        {
		    if( node.person.doc != null )
		    {
		        actionDisplayDocument( pageToURL(node.person.doc), true, true );
		    }
		    else
		    {
		        _history.removeAllElements();
        		doc.setDocument( new javax.swing.text.html.HTMLDocument() );
		    }
		}
		catch( Exception e )
		{
		    reportException( e );
		}

		setBusy( null );
    }


    private void actionListRowSelected( int index )
    {
        FamilyTreeTableRowModel row = _tableModel.getRow( index );
        if( row == null )
            return;

        try
        {
		    if( row.person.doc != null )
		    {
		        actionDisplayDocument( pageToURL(row.person.doc), true, true );
		    }
		    else
		    {
		        _history.removeAllElements();
        		doc.setDocument( new javax.swing.text.html.HTMLDocument() );
		    }
		}
		catch( Exception e )
		{
		    reportException( e );
		}

		setBusy( null );
    }


    public void actionHyperlinkActivated( URL url )
    {
        actionDisplayDocument( url, false, true );
    }


    public void actionHistoryBack()
    {
        if( _history.size() > 1 )
        {
            _history.removeElementAt( _history.size() -1 );
            URL url = (URL) _history.lastElement();

            actionDisplayDocument( url, false, false );
        }
    }


    private URL pageToURL( String page )
    {
        if( page.indexOf( '.' ) == -1 )
            page += ".html";

        page = "html/" + page;

        return ClassLoader.getSystemResource( page );
    }


    private void actionDisplayDocument( URL url, boolean clearHistory, boolean updateHistory )
    {
        if( url == null )
        {
            System.out.println( res.getString("WARNING_Attempt_to") );
            return;
        }

        // wrap jpegs in an ImageDocument
        if( (url.getFile() != null) && url.getFile().endsWith( ".jpg" ) )
        {
            try
            {
        	    doc.setDocument( new ImageDocument( url.toString() ) );
		        docscrollpane.setHorizontalScrollBarPolicy( javax.swing.ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED );

		        if( clearHistory )
		            _history.removeAllElements();

    		    if( updateHistory )
		            _history.addElement( url );
        	}
        	catch( Exception e )
        	{
        	    e.printStackTrace();
        	}
        	return;
        }

        // load normal pages as html
		try
		{
		    setBusy( res.getString("Loading") + url + res.getString("KEY") );
		    doc.setPage( url );
		    docscrollpane.setHorizontalScrollBarPolicy( javax.swing.ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER );

		    if( clearHistory )
		        _history.removeAllElements();

    		if( updateHistory )
		        _history.addElement( url );
		}
		catch( java.io.IOException e )
		{
		    //e.printStackTrace();
		    System.out.println( res.getString("IOException_") + e.getMessage() );
		}

		setBusy( null );
    }


    public void actionDocumentPrint()
    {
        DocumentPrintFrame dpf = new DocumentPrintFrame( (URL) _history.lastElement() );
        dpf.doPrint();
    }


    public void actionDocumentPopup( int x, int y )
    {
        _docPopupBackBtn.setEnabled( _history.size() > 1 );
        _docPopupPrintBtn.setEnabled( _history.size() > 0 );

        // Display the popup menu
        _docPopupMenu.show( doc, x, y );
    }


    // ActionListener for document display events
    private class DocumentActionListener
        implements ActionListener
    {
        private String _doc;

        public DocumentActionListener( String doc )  { _doc = doc; }

        public void actionPerformed( ActionEvent e )
        {
		    actionDisplayDocument( pageToURL(_doc), true, true );
        }
    }

    public void actionTreePopup( int x, int y )
    {
        TreePath path = tree.getPathForLocation( x, y );
        if( path == null ) return;

        FamilyTreeNode node = (FamilyTreeNode) path.getLastPathComponent();
        if( node == null ) return;

        // Build the tree popup menu
        JPopupMenu treePopupMenu = new JPopupMenu( res.getString("FamilyTree") );
        JMenuItem treePopupDescBtn = new JMenuItem( res.getString("Description") );
        treePopupMenu.add( treePopupDescBtn );
        treePopupDescBtn.setEnabled( node.person.doc != null );
        treePopupDescBtn.addActionListener( new DocumentActionListener( node.person.doc ) );

        if( node.person.getBornPage() != null )
        {
            JMenuItem treePopupBirthBtn = new JMenuItem( res.getString("Birth") );
            treePopupMenu.add( treePopupBirthBtn );
            treePopupBirthBtn.addActionListener( new DocumentActionListener( node.person.getBornPage() ) );
        }

        if( node.person.getChristenPage() != null )
        {
            JMenuItem treePopupChristenBtn = new JMenuItem( res.getString("Christening") );
            treePopupMenu.add( treePopupChristenBtn );
            treePopupChristenBtn.addActionListener( new DocumentActionListener( node.person.getChristenPage() ) );
        }

        if( node.person.getDiedPage() != null )
        {
            JMenuItem treePopupDeathBtn = new JMenuItem( res.getString("Death") );
            treePopupMenu.add( treePopupDeathBtn );
            treePopupDeathBtn.addActionListener( new DocumentActionListener( node.person.getDiedPage() ) );
        }

        if( node.person.getMarriedPage() != null )
        {
            JMenuItem treePopupMarriageBtn = new JMenuItem( res.getString("Marriage") );
            treePopupMenu.add( treePopupMarriageBtn );
            treePopupMarriageBtn.addActionListener( new DocumentActionListener( node.person.getMarriedPage() ) );
        }

        // Display the popup menu
        treePopupMenu.show( tree, x, y );
    }

	public void setVisible(boolean b)
	{
		if (b)
			setLocation(50, 50);
		super.setVisible(b);
	}

	static public void main(String args[])
	{
	    //UIManager.LookAndFeelInfo[] info = UIManager.getInstalledLookAndFeels();
	    //for( int n = 0; n < info.length; n++ )
	    //    System.out.println( info[n].getClassName() );

	    JWindow splash = new SplashWindow();
	    splash.setVisible(true);

	    try {
	        UIManager.setLookAndFeel( "com.sun.java.swing.plaf.windows.WindowsLookAndFeel" );
	    } catch( Exception e ) {
	        System.out.println( e.getMessage() );
	    }

		MainFrame frame = new MainFrame();

		if( args.length > 0 )
		    frame.setFilename( args[0] );

		try { Thread.sleep( 2000 ); } catch( InterruptedException e ) {}

		frame.setVisible(true);
		new Thread( frame ).start();

		splash.setVisible(false);
	}

	public void addNotify()
	{
		// Record the size of the window prior to calling parents addNotify.
		Dimension size = getSize();

		super.addNotify();

		if (frameSizeAdjusted)
			return;
		frameSizeAdjusted = true;

		// Adjust size of frame according to the insets and menu bar
		Insets insets = getInsets();
		javax.swing.JMenuBar menuBar = getRootPane().getJMenuBar();
		int menuBarHeight = 0;
		if (menuBar != null)
			menuBarHeight = menuBar.getPreferredSize().height;
		setSize(insets.left + insets.right + size.width, insets.top + insets.bottom + size.height + menuBarHeight);
	}

	// Used by addNotify
	boolean frameSizeAdjusted = false;

	//{{DECLARE_CONTROLS
	javax.swing.JSplitPane horizsplit = new javax.swing.JSplitPane();
	javax.swing.JScrollPane treescrollpane = new javax.swing.JScrollPane();
	javax.swing.JTree tree = new javax.swing.JTree();
	javax.swing.JSplitPane vertsplit = new javax.swing.JSplitPane();
	javax.swing.JScrollPane tablescrollpane = new javax.swing.JScrollPane();
	javax.swing.JTable table = new javax.swing.JTable();
	javax.swing.JScrollPane docscrollpane = new javax.swing.JScrollPane();
	javax.swing.JEditorPane doc = new javax.swing.JEditorPane();
	javax.swing.JTextField statusTxt = new javax.swing.JTextField();
	javax.swing.JMenuBar menubar = new javax.swing.JMenuBar();
	javax.swing.JMenu fileMenu = new javax.swing.JMenu();
	javax.swing.JMenuItem fileOpenBtn = new javax.swing.JMenuItem();
	javax.swing.JMenuItem filePageSetupBtn = new javax.swing.JMenuItem();
	javax.swing.JMenuItem fileExitBtn = new javax.swing.JMenuItem();
	javax.swing.JMenu viewMenu = new javax.swing.JMenu();
	javax.swing.JMenu showMenu = new javax.swing.JMenu();
	javax.swing.JCheckBoxMenuItem showParentsChk = new javax.swing.JCheckBoxMenuItem();
	javax.swing.JCheckBoxMenuItem showChildrenChk = new javax.swing.JCheckBoxMenuItem();
	javax.swing.JCheckBoxMenuItem showSiblingsChk = new javax.swing.JCheckBoxMenuItem();
	javax.swing.JCheckBoxMenuItem showAuntsUnclesChk = new javax.swing.JCheckBoxMenuItem();
	javax.swing.JCheckBoxMenuItem showSpousesChk = new javax.swing.JCheckBoxMenuItem();
	javax.swing.JCheckBoxMenuItem showCousinsChk = new javax.swing.JCheckBoxMenuItem();
	javax.swing.JMenu orientMenu = new javax.swing.JMenu();
	javax.swing.JRadioButtonMenuItem orientNormalRb = new javax.swing.JRadioButtonMenuItem();
	javax.swing.JRadioButtonMenuItem orientInvertedRb = new javax.swing.JRadioButtonMenuItem();
	javax.swing.JMenu viewSetRootMenu = new javax.swing.JMenu();
	javax.swing.JMenu toolsMenu = new javax.swing.JMenu();
	javax.swing.JMenuItem toolsDrawTreeBtn = new javax.swing.JMenuItem();
	javax.swing.JMenuItem toolsDrawAncestorsBtn = new javax.swing.JMenuItem();
	javax.swing.JMenuItem toolsDrawTimelineBtn = new javax.swing.JMenuItem();
	javax.swing.JMenu helpMenu = new javax.swing.JMenu();
	javax.swing.JMenuItem helpAboutBtn = new javax.swing.JMenuItem();
	//}}
    javax.swing.ButtonGroup viewSetRootGrp = new ButtonGroup();
    javax.swing.JRadioButtonMenuItem[] viewSetRootMenuItems;

    // Doc Popup Menu
    private JPopupMenu _docPopupMenu = new JPopupMenu( res.getString("FamilyTree") );
    private JMenuItem _docPopupBackBtn  = new JMenuItem( res.getString("Back") );
    private JMenuItem _docPopupPrintBtn = new JMenuItem( res.getString("Print_") );

	//{{DECLARE_MENUS
	//}}

	class SymAction implements java.awt.event.ActionListener
	{
		public void actionPerformed(java.awt.event.ActionEvent event)
		{
			Object object = event.getSource();
			if (object == fileExitBtn)
				fileExitBtn_actionPerformed(event);
		}
	}

	void fileExitBtn_actionPerformed(java.awt.event.ActionEvent event)
	{
		System.exit(0);
	}


	private void actionShowChanged()
	{
	    int flags = 0;

	    if( showParentsChk.isSelected() )       flags |= SHOW_PARENTS;
	    if( showSpousesChk.isSelected() )       flags |= SHOW_SPOUSES;
	    if( showChildrenChk.isSelected() )      flags |= SHOW_CHILDREN;
	    if( showSiblingsChk.isSelected() )      flags |= SHOW_SIBLINGS;
	    if( showAuntsUnclesChk.isSelected() )   flags |= SHOW_AUNTSUNCLES;
	    if( showCousinsChk.isSelected() )       flags |= SHOW_COUSINS;

	    _showFlags = flags;

	    if( _lastSelectedNode != null )
            actionTreeNodeSelected( _lastSelectedNode );
    }


    private void updateScreenState()
    {
        // show menu state
	    showParentsChk.setSelected( isShowing(SHOW_PARENTS) );
	    showSpousesChk.setSelected( isShowing(SHOW_SPOUSES) );
	    showChildrenChk.setSelected( isShowing(SHOW_CHILDREN) );
	    showSiblingsChk.setSelected( isShowing(SHOW_SIBLINGS) );
	    showAuntsUnclesChk.setSelected( isShowing(SHOW_AUNTSUNCLES) );
	    showCousinsChk.setSelected( isShowing(SHOW_COUSINS) );

        // orient menu state
	    orientNormalRb.setSelected( _orientation == FamilyTreeModel.CONVENTIONAL );
	    orientInvertedRb.setSelected( _orientation == FamilyTreeModel.INVERTED );

        // set root menu status
        for( int n = 0; n < viewSetRootMenuItems.length; n++ )
        {
            if( _orientation == FamilyTreeModel.CONVENTIONAL )
                viewSetRootMenuItems[n].setSelected( _selectedRoot == n );
            else
                viewSetRootMenuItems[n].setSelected( _selectedInvRoot == n );
        }

        boolean menusEnabled = ( _people != null );

        filePageSetupBtn.setEnabled( menusEnabled );
        viewMenu.setEnabled( menusEnabled );
        toolsMenu.setEnabled( menusEnabled );

        // set title bar
		this.setTitle( "Taurus FamilyTree - " + _filename );
    }


	private void actionOrientChanged()
	{
	    if( orientInvertedRb.isSelected() )
	    {
	        _orientation = FamilyTreeModel.INVERTED;
	        _root = _rootsInverted[_selectedInvRoot];
	        buildSetRootMenu( _rootsInverted );
	    }
	    else
	    {
	        _orientation = FamilyTreeModel.CONVENTIONAL;
	        _root = _rootsConventional[_selectedRoot];
	        buildSetRootMenu( _rootsConventional );
	    }

		tree.setModel( _model = new FamilyTreeModel( _people, _root, _orientation ) );
	}


	private boolean isShowing( int mask )
	{
	    return ((_showFlags & mask) == mask);
	}


	private void actionDrawTree()
	{
		try
		{
		    JFrame frame = new FTDiagramFrame( _rootsConventional[_selectedRoot] );
		    frame.setVisible( true );
		}
		catch( Exception e )
		{
		    e.printStackTrace();
		    System.out.println( e.toString() );
	    }
	}


	private void actionDrawAncestors()
	{
		try
		{
		    JFrame frame = new AncestorFrame( _rootsInverted[_selectedInvRoot] );
		    frame.setVisible( true );
		}
		catch( Exception e )
		{
		    e.printStackTrace();
		    System.out.println( e.toString() );
	    }
	}


	private void actionDrawTimeline()
	{
		try
		{
		    JFrame frame = new TLDiagramFrame( _people );
		    frame.setVisible( true );
		}
		catch( Exception e )
		{
		    e.printStackTrace();
		    System.out.println( e.toString() );
	    }
	}


	private void buildSetRootMenu( Person[] roots )
	{
        viewSetRootMenu.removeAll();
        viewSetRootGrp = new ButtonGroup();

        if( (roots == null) || (roots.length == 0) )
        {
            viewSetRootMenuItems = new JRadioButtonMenuItem[1];
            viewSetRootMenuItems[0] = new JRadioButtonMenuItem( "<None>" );
            viewSetRootMenu.add( viewSetRootMenuItems[0] );
            updateScreenState();
            return;
        }

        viewSetRootMenuItems = new JRadioButtonMenuItem[roots.length];

        ActionListener setRootAction = new ActionListener() {
		    public void actionPerformed( final ActionEvent event ) {
		        actionRootChanged();
		    }
	    };

        for( int n = 0; n < roots.length; n++ )
        {
            StringBuffer label = new StringBuffer();
            label.append( roots[n].getSurname().toUpperCase() );
            label.append( " " );
            label.append( roots[n].getForenames() );
            viewSetRootMenuItems[n] = new JRadioButtonMenuItem( label.toString() );
            viewSetRootMenu.add( viewSetRootMenuItems[n] );
            viewSetRootGrp.add( viewSetRootMenuItems[n] );
            viewSetRootMenuItems[n].addActionListener( setRootAction );
        }

        updateScreenState();
    }


    private void actionRootChanged()
    {
        for( int n = 0; n < viewSetRootMenuItems.length; n++ )
        {
            if( viewSetRootMenuItems[n].isSelected() )
            {
                if( _orientation == FamilyTreeModel.CONVENTIONAL )
                    _selectedRoot = n;
                else
                    _selectedInvRoot = n;
            }
        }

        actionOrientChanged();
    }

    private void actionOpen()
    {
        JFileChooser opendlg = new JFileChooser( "." );
        javax.swing.filechooser.FileFilter ftmlFilter;
        opendlg.addChoosableFileFilter( ftmlFilter = new javax.swing.filechooser.FileFilter() {
            public boolean accept( java.io.File f ) { return f.getName().endsWith( ".ftml" ) || f.getName().endsWith( ".xml" ) || f.isDirectory(); }
            public String getDescription() { return "FTML files (*.ftml, *.xml)"; }
        });
        opendlg.addChoosableFileFilter( new javax.swing.filechooser.FileFilter() {
            public boolean accept( java.io.File f ) { return f.getName().endsWith( ".csv" ) || f.isDirectory(); }
            public String getDescription() { return "CSV files (*.csv)"; }
        });
        opendlg.addChoosableFileFilter( new javax.swing.filechooser.FileFilter() {
            public boolean accept( java.io.File f ) { return f.getName().endsWith( ".ged" ) || f.isDirectory(); }
            public String getDescription() { return "Gedcom files (*.ged)"; }
        });
        opendlg.setFileFilter( ftmlFilter );

        if( opendlg.showOpenDialog( this ) == JFileChooser.APPROVE_OPTION )
        {
            java.io.File file = opendlg.getSelectedFile();
            setFilename( file.toString() );
            new Thread( this ).start();
        }
    }

    private void actionPageSetup()
    {
        new PageSetupDialog( this, false ).setVisible( true );
    }


    private void setBusy( String message )
    {
        if( message != null )
        {
	        statusTxt.setText( message );
	        statusTxt.repaint();
	        setCursor( new Cursor( Cursor.WAIT_CURSOR ) );
        }
        else
        {
	        setCursor( Cursor.getDefaultCursor() );
	        statusTxt.setText( "" );
	        statusTxt.repaint();
        }
    }


    private void reportException( Exception e )
    {
        StringWriter out = new StringWriter();
        PrintWriter pout = new PrintWriter( out );
        pout.println( e.toString() );
        e.printStackTrace( pout );
        JOptionPane.showMessageDialog( this, out.toString(), res.getString("AException_has"), JOptionPane.ERROR_MESSAGE );
        e.printStackTrace();
        System.out.println( e );
    }
}