///////////////////////////////////////////////////////////////////////////////
//
// Filename: DateUtil.java
//
// Author:   Robert Thornburrow
// Project:  FamilyTree
// Created:  20th May 2001
//
// Purpose:  Date manipulation utilities.
//
// (c) Copyright R.H.Thornburrow, 2001
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
//
///////////////////////////////////////////////////////////////////////////////

package com.tsac.familytree.util;

import java.util.*;
import java.text.*;

public class DateUtil
{
    public static DateFormat ddmmyyyy = new SimpleDateFormat( "dd/MM/yyyy" );
    public static DateFormat ddmmmyyyy = new SimpleDateFormat( "dd MMM yyyy" );
    public static DateFormat yyyy = new SimpleDateFormat( "yyyy" );
    
    public static DateBase toDateBase( String str )
    {
        DateBase ret = null;
        
        if( str != null )
        {
            try
            {
                // 31/12/2001
                Date date = ddmmyyyy.parse( str );
                Calendar cal = Calendar.getInstance();
                cal.setTime(date);
                ret = new AbsoluteDate( cal );
            }
            catch( Exception e )
            {
                try
                {
                    // 31 Dec 2001
                    Date date = ddmmmyyyy.parse( str );
                    Calendar cal = Calendar.getInstance();
                    cal.setTime(date);
                    ret = new AbsoluteDate( cal );
                }
                catch( Exception ex )
                {
                    try
                    {
                        // 2001
                        Date date = yyyy.parse( str );
                        Calendar cal = Calendar.getInstance();
                        cal.setTime(date);
                        ret = new ApproxDate( Integer.parseInt(str) );
                    }
                    catch( Exception ex2 )
                    {
                        System.out.println( "PARSE ERROR: " + e.getMessage() );
                    }
                }
            }
        }
        
        return ret;
    }
    
    public static String toString( Calendar cal )
    {
        if( cal != null )
            return ddmmyyyy.format( cal.getTime() );
            
        return " ";
    }
    
    public static int difference( Calendar start, Calendar end )
    {
        int years = end.get( Calendar.YEAR ) - start.get( Calendar.YEAR );
        int days = end.get( Calendar.DAY_OF_YEAR ) - start.get( Calendar.DAY_OF_YEAR );
        
        if( days >= 0 ) return years;
        
        return (years - 1);
    }
}